/**
 * Portalgun server-sent plugin made by Lorgan
 *
 * Portalgun sprite made by "Novally" http://novally.deviantart.com/art/Portal-Gun-Sprites-265796048
 */

globalvar PORTAL;
PORTAL = id;

// Constants
PORTAL.PLACE_PORTAL = 1;
PORTAL.DESTROY_PORTAL = 2;
PORTAL.PLACE_PEDESTAL = 3;
PORTAL.EQUIP = 4;
PORTAL.UNEQUIP = 5;
PORTAL.FULL_STATE = 6;

// Networking
PORTAL.sendBuffer = buffer_create();
PORTAL.portals = ds_list_create();
PORTAL.pedestals = ds_list_create();

// Sprites
PORTAL.PortalgunS = sprite_add(directory + "/Sprites/PortalgunS.png", 1, true, false, 7, 10);
PORTAL.PortalgunParticlesS = sprite_add(directory + "/Sprites/PortalgunParticlesS.png", 1, true, false, 7, 10);
PORTAL.PortalS = sprite_add(directory + "/Sprites/PortalS.png", 2, true, false, 32, 32);
PORTAL.ProjectileS = sprite_add(directory + "/Sprites/PortalProjectileS.png", 1, true, false, 4, 4);
PORTAL.PedestalS = sprite_add(directory + "/Sprites/PortalPedestalS.png", 1, true, false, 32, 40);
PORTAL.PortalCrosshairS = sprite_add(directory + "/Sprites/PortalCrosshairS.png", 2, true, false, 15, 15);
PORTAL.PortalKL = sprite_add(directory + "/Sprites/PortalKL.png", 2, true, false, 14, 9);

// Fix collision mask
sprite_collision_mask(PORTAL.PortalS, true, 0, -1, -1, -1, -1, 0, 0);

// Sound
PORTAL.ShotSnd = sound_add(directory + "/Sounds/PortalShot.wav", 0, true);

// Kill icon
createGlobalDamageSource("DAMAGE_SOURCE_PORTAL", PORTAL.PortalKL);

// scripts
execute_file(directory + "/scripts.gml");
execute_file(directory + "/portal.gml");
execute_file(directory + "/chat_ext.gml");

execute_file(directory + "/lorganbase.gml");
ds_list_add(lorganbase_plugins, "portal");

// Load pedestals on the map.
var entityCount, pedestalEntity, map;
entityCount = ds_list_size(global.entities);
pedestalEntity = addEntity("portal_pedestal", -1, "{color:ff7f00,respawnTime:0}", PORTAL.Pedestal, -1, -1, -1, -1);

// We did not add the entity which means the Portal Builder was loaded. In that case we need to fix the object that gets spawned.
if (ds_list_size(global.entities) == entityCount) {
    map = ds_list_find_value(global.entityData, pedestalEntity);
    ds_map_replace(map, "object", PORTAL.Pedestal);
}

// Handle networking and input
object_event_add(PlayerControl, ev_step, ev_step_begin, '
    var receiveBuffer;
    while(PluginPacketGetBuffer(PORTAL.packetID) != -1) {
        receiveBuffer = PluginPacketGetBuffer(PORTAL.packetID);
        if (global.isHost == false) {
            var action;
            action = read_ubyte(receiveBuffer);
            switch(action) {
                case PORTAL.PLACE_PORTAL:
                    execute_string(PORTAL.PlacePortal, read_ushort(receiveBuffer) / 5, read_ushort(receiveBuffer) / 5, read_ubyte(receiveBuffer), read_ubyte(receiveBuffer) * 2, make_color_rgb(read_ubyte(receiveBuffer), read_ubyte(receiveBuffer), read_ubyte(receiveBuffer)));
                    break;
                case PORTAL.DESTROY_PORTAL:
                    execute_string(PORTAL.DestroyPortal, read_ubyte(receiveBuffer));
                    break;
                case PORTAL.PLACE_PEDESTAL:
                    execute_string(PORTAL.PlacePedestal, read_ushort(receiveBuffer) / 5, read_ushort(receiveBuffer) / 5, make_color_rgb(read_ubyte(receiveBuffer), read_ubyte(receiveBuffer), read_ubyte(receiveBuffer)), read_ushort(receiveBuffer));
                    break;
                case PORTAL.EQUIP:
                    execute_string(PORTAL.Equip, read_ubyte(receiveBuffer), read_ubyte(receiveBuffer), make_color_rgb(read_ubyte(receiveBuffer), read_ubyte(receiveBuffer), read_ubyte(receiveBuffer)));
                    break;
                case PORTAL.UNEQUIP:
                    execute_string(PORTAL.Unequip, read_ubyte(receiveBuffer), read_ubyte(receiveBuffer));
                    break;
                case PORTAL.FULL_STATE:
                    execute_string(PORTAL.ReadFullState, receiveBuffer);
                    break;
            }
        } else {
            var player, foundPedestal;
            player = PluginPacketGetPlayer(PORTAL.packetID);
            if (player == noone) { // WHY IS THIS NOONE THIS IS STUPID
                player = global.myself;
            }

            if (player != -1) {
                if (player.object != -1) {
                    foundPedestal = false;

                    with(PORTAL.Pedestal) {
                        if (distance_to_point(player.object.x, player.object.y) < 30) {
                            if (hasPortalgun == true) {
                                execute_string(PORTAL.SendEquip, player, id, color);
                                foundPedestal = true;
                            } else if (player.object.currentWeapon.object_index == PORTAL.Portalgun) {
                                execute_string(PORTAL.SendUnequip, player, id);
                                foundPedestal = true;
                            }
                        }

                        if (foundPedestal == true) {
                            break;
                        }
                    }

                    if (foundPedestal == false && player.object.currentWeapon.object_index == PORTAL.Portalgun) {
                        with(player.object.currentWeapon.primary) {
                            execute_string(PORTAL.SendDestroyPortal, id);
                        }
                        with(player.object.currentWeapon.secondary) {
                            execute_string(PORTAL.SendDestroyPortal, id);
                        }
                    }
                }
            }
        }

        PluginPacketPop(PORTAL.packetID);
    }

    var kickOpen;
    kickOpen = false
    if (instance_exists(ScoreTableController))
        if (ScoreTableController.showadmin)
            kickOpen = true;

    if (instance_exists(MenuController) || instance_exists(TeamSelectController) || instance_exists(ClassSelectController) || kickOpen == true || global.myself.object == -1) {
        exit;
    }

    if (keyboard_check_pressed(ord("Q")) == true) {
        if (global.isHost == true) {
            PluginPacketSendTo(PORTAL.packetID, PORTAL.sendBuffer, global.myself);
        } else {
            PluginPacketSend(PORTAL.packetID, PORTAL.sendBuffer);
        }

        buffer_clear(PORTAL.sendBuffer);
    }
');

// Send initial state to the client.
object_event_add(Player, ev_create, 0, '
    if (global.isHost) {
        alarm[1] = 1;
    }
');
object_event_add(Player, ev_alarm, 1, '
    execute_string(PORTAL.SendFullState);
');

// Character
object_event_add(Character, ev_create, 0, '
    lastPortal = noone;
    sourcePedestal = noone;
');

object_event_add(Character, ev_step, ev_step_begin, '
    oldHspeed = hspeed;
');

object_event_add(Character, ev_step, ev_step_normal, '
    if (moveStatus == 3) {
        // conteract the fast slowing down when coming out of a portal.
        if(abs(hspeed) > basemaxspeed * 2 or ((keyState|pressedKeys) & $60 and abs(hspeed) < basemaxspeed)) {
            hspeed *= delta_mult_skip(baseFriction) ;

        } else {
            hspeed *= delta_mult_skip(frictionFactor);
        }
    }

    if (place_meeting(x, y, PORTAL.Portal) == false) {
        lastPortal = noone;
    }
');

object_event_add(Character, ev_destroy, 0, '
    with(sourcePedestal) {
        hasPortalgun = true;
        alarm[0] = -1;
    }
');

object_event_add(Character, ev_collision, PORTAL.Portal, '
    // Dont make it /too/ easy.
    if (intel == true) {
        exit;
    }

    // Exit the portal before being teleported again.
    if (other.id == lastPortal) {
        exit;
    }

    // Look for the (first) complementary portal.
    var complementaryColor, pair;
    complementaryColor = make_color_rgb(255 - color_get_red(other.color), 255 - color_get_green(other.color), 255 - color_get_blue(other.color));
    pair = noone;

    with(PORTAL.Portal) {
        if (color == complementaryColor) {
            pair = id;
            break;
        }
    }

    // Teleport if close enough.
    if (pair != noone) {
        if ((other.image_index == 0 && abs(y - other.y) < 25 && abs(hspeed) > abs(vspeed) - 5) || // Wall portal
            (other.image_index == 1 && abs(x - other.x) < 15)) { // Floor portal

            // This is a hack but who cares.
            var offset;
            switch(pair.direction) {
                case 0:
                    offset = 10;
                    break;
                case 90:
                    offset = 32;
                    break;
                case 180:
                    offset = 10;
                    break;
                case 270:
                    offset = 32;
                    break;
            }

            // Shoot out of the other portal with your current velocity.
            var tmpSpeed;
            tmpSpeed = speed;
            hspeed = lengthdir_x(tmpSpeed, pair.direction);
            vspeed = lengthdir_y(tmpSpeed, pair.direction);

            // Dont spawn in the wall.
            x = pair.x + lengthdir_x(offset, pair.direction);
            y = pair.y + lengthdir_y(offset, pair.direction);
            lastPortal = pair;

            moveStatus = 3;
        }
    }
');

object_event_add(Character, ev_other, ev_user12, '
    if(global.updateType == FULL_UPDATE) {
        if (currentWeapon.object_index == PORTAL.Portalgun) {
            write_ubyte(global.serializeBuffer, 1);
            write_ubyte(global.serializeBuffer, currentWeapon.colorR);
            write_ubyte(global.serializeBuffer, currentWeapon.colorG);
            write_ubyte(global.serializeBuffer, currentWeapon.colorB);
        } else {
            write_ubyte(global.serializeBuffer, 0);
        }
    }
');

object_event_add(Character, ev_other, ev_user13, '
    if(global.updateType == FULL_UPDATE) {
        receiveCompleteMessage(global.serverSocket, 1, global.deserializeBuffer);
        if (read_ubyte(global.deserializeBuffer) == 1) {
            with(player.object.currentWeapon) {
                instance_destroy();
            }

            global.paramOwner = player.object;
            player.object.currentWeapon = instance_create(x, y, PORTAL.Portalgun);
            global.paramOwner = noone;

            receiveCompleteMessage(global.serverSocket, 3, global.deserializeBuffer);
            player.object.currentWeapon.colorR = read_ubyte(global.deserializeBuffer);
            player.object.currentWeapon.colorG = read_ubyte(global.deserializeBuffer);
            player.object.currentWeapon.colorB = read_ubyte(global.deserializeBuffer);
        }
    }
');

// Teleport bullets and stuff
object_event_add(Shot, ev_collision, PORTAL.Portal, '
    execute_string(PORTAL.TeleportProjectile, id, other.id);
');

object_event_add(Needle, ev_collision, PORTAL.Portal, '
    execute_string(PORTAL.TeleportProjectile, id, other.id);
');

object_event_add(Mine, ev_collision, PORTAL.Portal, '
    //if (variable_local_exists("teleported") == false) {
        stickied = false;
    //}

    execute_string(PORTAL.TeleportProjectile, id, other.id, true);
');

object_event_add(BurningProjectile, ev_collision, PORTAL.Portal, '
    execute_string(PORTAL.TeleportProjectile, id, other.id);
');

object_event_add(Rocket, ev_collision, PORTAL.Portal, '
    execute_string(PORTAL.TeleportProjectile, id, other.id);
');

object_event_add(Sentry, ev_collision, PORTAL.Portal, '
    execute_string(PORTAL.TeleportProjectile, id, other.id, true);

    // Make sure to sync the new location.
    oldVspeed = 999;
');

object_event_add(Sentry, ev_draw, 0, '
    if(vspeed == 0 && built == 1) {
        image_index = 11;
    } else if (currentWeapon != -1) {
        // Dont shoot while parashuting.
        currentWeapon.readyToShoot = false;
        currentWeapon.alarm[0] = currentWeapon.refireTime / global.delta_factor;
    }
');

object_event_add(Sentry, ev_other, ev_user2, '
    if (vspeed == 0) {
        hspeed = 0;
    }
');

// Use blood as portal particles.
object_event_add(BloodDrop, ev_create, 0, '
    color = c_white;
');

object_event_add(BloodDrop, ev_draw, 0, '
    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, color, image_alpha);
');