// Pedestal
PORTAL.Pedestal = object_add();
object_set_sprite(PORTAL.Pedestal, PORTAL.PedestalS);

object_event_add(PORTAL.Pedestal, ev_create, 0, '
    // Align with mappixels.
    y += 1;

    color = make_color_rgb(255, 127, 0);
    defaultColor = color;
    hasPortalgun = true;
    respawnTime = 0;

    ownedByMap = true; // Was this pedestal placed by the map?
    ds_list_add(PORTAL.pedestals, id);
');

object_event_add(PORTAL.Pedestal, ev_alarm, 0, '
    hasPortalgun = true;
');

object_event_add(PORTAL.Pedestal, ev_destroy, 0, '
    ds_list_delete(PORTAL.pedestals, ds_list_find_index(PORTAL.pedestals, id));
');

object_event_add(PORTAL.Pedestal, ev_other, ev_user1, '
    color = readProperty(other.properties, "color", HEX, color);
    respawnTime = readProperty(other.properties, "respawnTime", REAL, respawnTime);

    defaultColor = color;
');

object_event_add(PORTAL.Pedestal, ev_draw, 0, '
    draw_sprite_ext(sprite_index, 0, x, y, image_xscale, image_yscale, image_angle, c_white, image_alpha);

    if (hasPortalgun == true) {
        draw_sprite_ext(PORTAL.PortalgunS, 0, x - 10 * image_xscale, y - 4, image_xscale, image_yscale, image_angle, c_white, image_alpha);
    }

    if (distance_to_point(view_xview + view_wview / 2, view_yview + view_hview / 2) < 30) {
        var text;
        text = "Press " + chr(39) + "Q" + chr(39);

        draw_set_alpha(1);
        draw_set_valign(fa_bottom);
        draw_set_halign(fa_center);

        draw_set_color(c_black);
        draw_text(x, y - 20, text);

        draw_set_color(c_white);
        draw_text(x - 2, y - 22, text);
    }
');

// Projectile
PORTAL.Projectile = object_add();
object_set_parent(PORTAL.Projectile, BladeB); // Behave like a blade because it flies straight.
object_set_sprite(PORTAL.Projectile, PORTAL.ProjectileS);

object_event_add(PORTAL.Projectile, ev_create, 0, '
    color = c_white;

    hitDamage = 10;
    baseDamage = hitDamage;
');

object_event_add(PORTAL.Projectile, ev_step, ev_step_begin, '');
object_event_add(PORTAL.Projectile, ev_step, ev_step_end, '');

object_event_add(PORTAL.Projectile, ev_destroy, 0, '
    gunSetSolids();
    var portal, dir, offset;
    portal = noone;
    dir = 1;
    offset = 2;

    // Only stick to the wallmask.
    // There is some 60 fps inconsistency here, a bit to the position seems to be required in 30 fps mode.
    if (place_meeting(x + lengthdir_x(offset, direction), y + lengthdir_y(offset, direction), Obstacle) &&
       !place_meeting(x + lengthdir_x(offset, direction), y + lengthdir_y(offset, direction), PORTAL.Portal)) {

        really_move_contact_solid(direction, speed);

        if (global.isHost == true) {
            if (place_free(x, y + 25) && place_free(x, y - 25)) {
                if (direction > 90 && direction < 270) {
                    dir *= -1;
                }

                if (!place_free(x + offset * dir, y + 25) && !place_free(x + offset * dir, y - 25)) {
                    portal = execute_string(PORTAL.SendPlacePortal, x, y, 0, 90 + dir * 90, color);
                }
            } else if (place_free(x + 15, y) && place_free(x - 15, y)) {
                if (direction > 0 && direction < 180) {
                    dir *= -1;
                }

                if (!place_free(x + 15, y + offset * dir) && !place_free(x - 15, y + offset * dir)) {
                    portal = execute_string(PORTAL.SendPlacePortal, x, y, 1, dir * 90, color);
                }
            }

            // Make sure only one portal exists.
            if (portal != noone) {
                with(PORTAL.Portal) {
                    if (color == other.color && id != portal) {
                        execute_string(PORTAL.SendDestroyPortal, id);
                    }
                }
            }
        }
    }

    execute_string(PORTAL.SpawnParticles, x, y, direction + 180, color);
    gunUnsetSolids();
');

object_event_add(PORTAL.Projectile, ev_draw, 0, '
    draw_sprite_ext(sprite_index, 0, x, y, image_xscale, image_yscale, image_angle, color, image_alpha);
');

// Portal
PORTAL.Portal = object_add();
object_set_sprite(PORTAL.Portal, PORTAL.PortalS);

object_event_add(PORTAL.Portal, ev_create, 0, '
    image_speed = 0;
    color = c_white;

    ds_list_add(PORTAL.portals, id);
');

object_event_add(PORTAL.Portal, ev_destroy, 0, '
    ds_list_delete(PORTAL.portals, ds_list_find_index(PORTAL.portals, id));
    execute_string(PORTAL.SpawnParticles, x, y, direction, color);
');

object_event_add(PORTAL.Portal, ev_draw, 0, '
    draw_sprite_ext(sprite_index, image_index, x, y, image_xscale, image_yscale, image_angle, color, image_alpha);
');

// Gun
PORTAL.Portalgun = object_add();
object_set_parent(PORTAL.Portalgun, Blade); // Blade is still pretty much the best choice.
object_set_sprite(PORTAL.Portalgun, PORTAL.PortalgunS);
object_set_depth(PORTAL.Portalgun, -1);

object_event_add(PORTAL.Portalgun, ev_create, 0, '
    event_inherited();

    xoffset = 0;
    yoffset = 2;

    refireTime = 40;
    colorR = 255;
    colorG = 127;
    colorB = 0;
    lastShotColor = 0;

    primary = noone;
    secondary = noone;

    normalSprite = PORTAL.PortalgunS;
    recoilSprite = PORTAL.PortalgunS;
    reloadSprite = -1;

    recoilTime = 1;
    recoilAnimLength = sprite_get_number(recoilSprite) / 2;
    recoilImageSpeed = recoilAnimLength/recoilTime;

    // Prevent crashes...
    t = 0;
    unscopedDamage = 0;
    baseDamage = 0;
    uberCharge = 0;
    lobbed = 0;
    readyToStab = true; // True or Spy cant uncloak!
');

object_event_add(PORTAL.Portalgun, ev_step, ev_step_end, '
    event_inherited();

    // Do not want.
    if (ownerPlayer == global.myself) {
        with(AmmoCounter) {
            instance_destroy();
        }

        with(StickyCounter) {
            instance_destroy();
        }
    }
');

// Shoot orange portal
object_event_add(PORTAL.Portalgun, ev_other, ev_user1, '
    if (readyToShoot && !owner.cloak) {
        playsound(x, y, PORTAL.ShotSnd);

        var shot;
        randomize();
        shot = createShot(x + lengthdir_x(8, owner.aimDirection), y + lengthdir_y(8, owner.aimDirection), PORTAL.Projectile, DAMAGE_SOURCE_PORTAL, owner.aimDirection, 17);
        shot.gun = id;
        shot.color = make_color_rgb(colorR, colorG, colorB);
        shot.primary = true;

        justShot = true;
        readyToShoot = false;
        alarm[0] = refireTime / global.delta_factor;
        lastShotColor = shot.color;
    }
');

// Shoot blue portal
object_event_add(PORTAL.Portalgun, ev_other, ev_user2, '
    if (readyToShoot && !owner.cloak) {
        playsound(x, y, PORTAL.ShotSnd);

        var shot;
        randomize();
        shot = createShot(x + lengthdir_x(8, owner.aimDirection), y + lengthdir_y(8, owner.aimDirection), PORTAL.Projectile, DAMAGE_SOURCE_PORTAL, owner.aimDirection, 17);
        shot.gun = id;
        shot.color = make_color_rgb(255 - colorR, 255 - colorG, 255 - colorB);
        shot.primary = false;

        justShot = true;
        readyToShoot = false;
        alarm[0] = refireTime / global.delta_factor;
        lastShotColor = shot.color;
    }
');

// Simulate the original weapons to prevent clients from locking up.
object_event_add(PORTAL.Portalgun, ev_other, ev_user12, "
    switch(ownerPlayer.class) {
        case CLASS_MEDIC:
            write_ubyte(global.serializeBuffer, 255);
        break;
        case CLASS_DEMOMAN:
            event_inherited();
            write_ubyte(global.serializeBuffer, 0);
        break;
        default:
            event_inherited();
        break;
    }
");

object_event_add(PORTAL.Portalgun, ev_other, ev_user14, '
    if (global.isHost == true) {
        with(primary) {
            execute_string(PORTAL.SendDestroyPortal, id);
        }

        with(secondary) {
            execute_string(PORTAL.SendDestroyPortal, id);
        }
    }
');

object_event_add(PORTAL.Portalgun, ev_draw, 0, '
    if (distance_to_point(view_xview + view_wview/2, view_yview + view_hview/2) > 800) {
        exit;
    }

    if (!owner.invisible and !owner.taunting and !owner.omnomnomnom and !owner.player.humiliated) {
        if (!owner.cloak)
            image_alpha = power(owner.cloakAlpha, 0.5);
        else
            image_alpha = power(owner.cloakAlpha, 2);

        draw_sprite_ext(sprite_index, 0, round(x+xoffset*image_xscale), round(y+yoffset) + owner.equipmentOffset, image_xscale, image_yscale, image_angle, c_white, image_alpha);

        if (owner.ubered) {
            if (owner.team == TEAM_RED)
                ubercolour = c_red;
            else if (owner.team == TEAM_BLUE)
                ubercolour = c_blue;
            draw_sprite_ext(sprite_index, 0, round(x+xoffset*image_xscale), round(y+yoffset) + owner.equipmentOffset, image_xscale, image_yscale, image_angle, ubercolour, 0.7*image_alpha);
        }

        if (lastShotColor != 0) {
            draw_sprite_ext(PORTAL.PortalgunParticlesS, 0, round(x+xoffset*image_xscale), round(y+yoffset) + owner.equipmentOffset, image_xscale, image_yscale, image_angle, lastShotColor, image_alpha);
        }
    }

    if (instance_exists(primary) == true) {
        draw_sprite_ext(PORTAL.PortalCrosshairS, 0, mouse_x, mouse_y, 1, 1, 0, primary.color, 1);
    }

    if (instance_exists(secondary) == true) {
        draw_sprite_ext(PORTAL.PortalCrosshairS, 1, mouse_x, mouse_y, 1, 1, 0, secondary.color, 1);
    }
');